c Uso: 
c - Eseguire: php socialGolfer.php <P> <G> <W>
c
c Esecuzione con zChaff su istanza data e successiva traduzione del modello trovato in forma intellegibile
c php socialGolfer.php <P> <G> <W> > cnf.txt ; zchaff cnf.txt | php zchaff2SocialGolfer.php cnf.txt
c
c Social Golfer to SAT compiler (the most straightforward encoding, just a simple exercise for students)
c Author: Toni Mancini - Sapienza University of Rome
c Date: July 16, 2008
c
<?php

	$DEBUG = FALSE;

	$P = $argv[1];
	$G = $argv[2];
	$W = $argv[3];

	if ($P == NULL || !is_numeric($P) || $P <= 0 ||
		$G == NULL || !is_numeric($G) || $G <= 0 ||
		$W == NULL || !is_numeric($W) || $W <= 0
	) {
		die("Syntax: php ".$argv[0]." <P> <G> <W>\n P > 0: Number of players\n G > 0: Number of groups\n W > 0: Number of weeks");
	};
	
	$S = $P/$G;
	if (!is_integer($S)) {
		die("Error: P/G (group size) must be integer");
		
	}
	

	$CLAUSES = array();
	$NCLAUSES = 0;
	$NVARS = 0;
	
	// LP = { PLAYS_p_g_w | p in 1..P, g in 1..G, w in 1..W }
	//        PLAYS_p_g_w represents player p playing in group g on week w
	
	$dictionary = NULL; // will be a 4D matrix: 
		// dictionary['PLAYS'][$p][$g][$w] <--> PLAYS(p,g,w)  and
		// dictionary['ALL_IN_GROUP'][$g][$w][<p1...ps>]  <--> ALL_IN_GROUP(g,w, <p1..ps>) --aux variables, cf. below
	
	
?>
c
c Problem description:
c In a golf club there are P social golfers who play once a week in G groups of the same size 
c (S=P/G, assumed integer). 
c The problem amounts to find a schedule for W weeks, such that no two golfers play in the 
c same group more than once.
c
<?php
	
	echo("c Number of players: $P\n");
	echo("c Number of groups: $G\n");
	echo("c Number of weeks: $W\n");	

	echo("c\nc Variables:\nc - plays(player, group, week)\n");
	echo("c - all_in_group(group, week, <players>) (auxiliary variables added to model the same-size constraint more easily (not requested at exam)\n");



	if ($DEBUG) addComment("c Each player plays in at least one group on every week");
	// AND_{w in 1..W) AND_{p in 1..P} OR_{g in 1..G} PLAYS_p,g,w
	for ($w=1; $w<=$W; $w++) {
		for ($p=1; $p<=$P; $p++) {	
			$clause = "";
			for ($g=1; $g<=$G; $g++) {
				$clause .= plays2dimacs($p,$g,$w) . " ";
			}
			addClause($clause);
		}
	}
	
	if ($DEBUG) addComment("c\nc\nc Each player play in at most one group on every week");
	// AND_{w in 1..W) AND_{p in 1..P} AND_{g1 in 1..G} AND_{g2 in 1..G, g2 != g1) (!PLAYS_p_g1_w || !PLAYS_p_g2_w)	
	for ($w=1; $w<=$W; $w++) {
		for ($p=1; $p<=$P; $p++) {	
			for ($g1=1; $g1<=$G; $g1++) {	
				for ($g2=$g1+1; $g2<=$G; $g2++) {	// Avoiding repeating clauses			
					addClause("-".plays2dimacs($p,$g1,$w) . " -" . plays2dimacs($p,$g2,$w));
				}
			}
		}
	}
	
	if ($DEBUG) addComment("c\nc\nc Two players play in the same group at most once");	
	// AND_{p1,p2 in 1..P, p1!=p2} AND_{g1,g2 in 1..G} AND_{w1,w2 in 1..W, w1!=w2}
	//  (PLAYS_p1,g1,w1 & PLAYS_p2,g1,w1 --> ! ( PLAYS_p1,g2,w2 & PLAYS_p2,g2,w2 )
	for ($p1=1; $p1<=$P; $p1++) {	
		for ($p2=$p1+1; $p2<=$P; $p2++) {	
			for ($g1=1; $g1<=$G; $g1++) {	
				for ($g2=1; $g2<=$G; $g2++) {	
					for ($w1=1; $w1<=$W; $w1++) {	
						for ($w2=$w1+1; $w2<=$W; $w2++) {	
							addClause("-".plays2dimacs($p1,$g1,$w1) . " -" . plays2dimacs($p2,$g1,$w1) .
									 " -".plays2dimacs($p1,$g2,$w2) . " -" . plays2dimacs($p2,$g2,$w2)
							);
						}
					}
				}
			}
		}
	}

	if ($DEBUG) addComment("c\nc\nc For every week, all groups have the same size $S");	
	/* AND_{g in 1..G} AND_{w in 1..W}  | { p in 1..P :  PLAYS_p,g,w == true } | = S
	   This is more difficult to encode as a set of clauses (not requested at exam)
	   Subformula | { p in 1..P :  PLAYS_p,g,w == true } | = S
	   can be encoded as OR_{ (p1...ps) s.t. p1..ps are all distinct and in 1..P } ( AND_{j in 1..s} PLAYS_pj_g_w ), hence in DNF. We have to convert this into CNF.
			
  	   An idea could be that of introducing new propositional (redundant) variables, to ease conversion:
	   for every g, w, and **unordered** S-tuple <p1...ps> (pj in 1..P) we add new variable 
	   ALL_IN_GROUP_g_w_<p1...ps> and replace formula with:
		 
	   AND_{g in 1..G} AND_{w in 1..W}
	         OR_{ <p1...ps> s.t. p1..ps are all distinct and in 1..P } ALL_IN_GROUP_g_w_<p1...ps>.
		 
	   AND 
	   
	   AND_{g in 1..G} AND_{w in 1..W} AND_{ <p1...ps> s.t. p1..ps are all distinct and in 1..P }
	       ( ALL_IN_GROUP_g_w_<p1...ps> <-> (PLAYS_p1_g_w & ... & PLAYS_ps_g_w) & AND_{all other players p'} -PLAYS_p'_g_w)
			 
			 
			 
	   The first part is already in CNF, while the second one may be rewritten as:
	   
	   AND_{g in 1..G} AND_{w in 1..W} AND_{ <p1...ps> s.t. p1..ps are all distinct and in 1..P }
           ( ALL_IN_GROUP_g_w_<p1...ps> -> (PLAYS_p1_g_w & ... & PLAYS_ps_g_w & AND_{all other players p'} -PLAYS_p'_g_w) 
       AND
       AND_{g in 1..G} AND_{w in 1..W} AND_{ <p1...ps> s.t. p1..ps are all distinct and in 1..P }
	       ( (PLAYS_p1_g_w & ... & PLAYS_ps_g_w & AND_{all other players p'} -PLAYS_p'_g_w) -> ALL_IN_GROUP_g_w_<p1...ps> )
			 
 	   hence as:
 	   
  	   AND_{g in 1..G} AND_{w in 1..W} AND_{ <p1...ps> s.t. p1..ps are all distinct and in 1..P }
  	   		!ALL_IN_GROUP_g_w_<p1...ps> || PLAYS_p1_g_w 
  	   		AND
  	   		...
  	   		AND
  	   		!ALL_IN_GROUP_g_w_<p1...ps> || PLAYS_ps_g_w 
  	   		AND
  	   		AND_{all other players p'}  !ALL_IN_GROUP_g_w_<p1...ps> || !PLAYS_p'_g_w 

	   AND
  	   
       AND_{g in 1..G} AND_{w in 1..W} AND_{ <p1...ps> s.t. p1..ps are all distinct and in 1..P }
			 !PLAYS_p1_g_w ||  !PLAYS_ps_g_w || OR_{all other players p'} PLAYS_p'_g_w || ALL_IN_GROUP_g_w_<p1...ps>.




	   The whole formula in CNF is thus:
	   
	   (1) AND_{g in 1..G} AND_{w in 1..W}
	         OR_{ <p1...ps> s.t. p1..ps are all distinct and in 1..P } ALL_IN_GROUP_g_w_<p1...ps>.
		 
	   AND 	   

  	   (2) AND_{g in 1..G} AND_{w in 1..W} AND_{ <p1...ps> s.t. p1..ps are all distinct and in 1..P }
  	   		!ALL_IN_GROUP_g_w_<p1...ps> || PLAYS_p1_g_w 
  	   		AND
  	   		...
  	   		AND
  	   		!ALL_IN_GROUP_g_w_<p1...ps> || PLAYS_ps_g_w 
  	   		AND
  	   		AND_{all other players p'}  !ALL_IN_GROUP_g_w_<p1...ps> || !PLAYS_p'_g_w 

	   AND
  	   
       (3) AND_{g in 1..G} AND_{w in 1..W} AND_{ <p1...ps> s.t. p1..ps are all distinct and in 1..P }
			 !PLAYS_p1_g_w ||  !PLAYS_ps_g_w || OR_{all other players p'} PLAYS_p'_g_w || ALL_IN_GROUP_g_w_<p1...ps>.

		
	   The only remaining point is to compute all possible **unordered** S-tuples <p1...ps>.
	   We encode any such tuple as an **ordered** array of S elements (array ordering allows us to 
	   avoid considering two tuples with the same elements).
	*/


	/* (1) AND_{g in 1..G} AND_{w in 1..W}
	         OR_{ <p1...ps> s.t. p1..ps are all distinct and in 1..P } ALL_IN_GROUP_g_w_<p1...ps>.
	*/
	if ($DEBUG) addComment("c\nc    --- (1)");		
	for ($g=1; $g<=$G; $g++) {	
		for ($w=1; $w<=$W; $w++) {	
			$clause = "";
			$tuple = firstTuple();
			do {
				$clause .= allingroup2dimacs($g,$w,$tuple) . " ";
				$tuple = nextTuple($tuple);		
			} while ($tuple !== NULL);
			addClause($clause);
		}
	}

	/*
  	   (2) AND_{g in 1..G} AND_{w in 1..W} AND_{ <p1...ps> s.t. p1..ps are all distinct and in 1..P }
  	   		!ALL_IN_GROUP_g_w_<p1...ps> || PLAYS_p1_g_w 
  	   		AND
  	   		...
  	   		AND
  	   		!ALL_IN_GROUP_g_w_<p1...ps> || PLAYS_ps_g_w 
  	   		AND
  	   		AND_{all other players p'}  !ALL_IN_GROUP_g_w_<p1...ps> || !PLAYS_p'_g_w 
	*/
	if ($DEBUG) addComment("c\nc    --- (2)");		
	for ($g=1; $g<=$G; $g++) {	
		for ($w=1; $w<=$W; $w++) {	
			$tuple = firstTuple();
			do {
				for($p=1; $p<=$P; $p++) {
					$clause = "-" . allingroup2dimacs($g,$w,$tuple). " ";
					if (in_array($p, $tuple)) {
						$clause .= plays2dimacs($p,$g,$w);
					} else {
						$clause .= "-" . plays2dimacs($p,$g,$w);					
					}
					addClause($clause);
				}			
				$tuple = nextTuple($tuple);		
			} while ($tuple !== NULL);
		}
	}
	

	/*
       (3) AND_{g in 1..G} AND_{w in 1..W} AND_{ <p1...ps> s.t. p1..ps are all distinct and in 1..P }
			 !PLAYS_p1_g_w ||  !PLAYS_ps_g_w || OR_{all other players p'} PLAYS_p'_g_w || ALL_IN_GROUP_g_w_<p1...ps>.
	*/
	if ($DEBUG) addComment("c\nc    --- (3)");		
	for ($g=1; $g<=$G; $g++) {	
		for ($w=1; $w<=$W; $w++) {	
			$tuple = firstTuple();
			do {
				$clause = "";
				for($p=1; $p<=$P; $p++) {
					if (in_array($p, $tuple)) {
						$clause .= "-" . plays2dimacs($p,$g,$w) . " ";
					} else {
						$clause .= plays2dimacs($p,$g,$w) . " ";					
					}
				}	
				$clause .= allingroup2dimacs($g,$w,$tuple);
				addClause($clause);				
				$tuple = nextTuple($tuple);		
			} while ($tuple !== NULL);
		}
	}




	/* Additional constraints, to improve efficiency.
	   Simple symmetry-breaking: since group names have no actual meaning, we can fix player 1 to play always in group 1
	   (much more interesting symmetry-breaking constraints can be added, cf., e.g., 
	   http://www.dis.uniroma1.it/~tmancini/index.php?currItem=research.publications.webappendices.csplib2x.problemDetails&problemid=010
	   ).
	   
	   Adding symmetry-breaking constraints preserves satisfiability of the problem (new problem satisfiable iff original one is satisfiable), but some solutions (symmetric ones) are lost.  
	   Especially advantageous in case of unsatisfiable instances (search space to be "entirely" explored is much smaller).
	   
	   More info at url above (see paper).
	*/
	
	if ($DEBUG) addComment("c Symmetry-breaking (for efficiency only): player 1 plays always in group 1");
	for($w=1; $w<=$W; $w++) {
		addClause( plays2dimacs(1, 1, $w) );
	}



	// Print out dictionary
	echoDictionary();
	
	// Print out a list of structural properties of variables (useful for ad-hoc solvers)
	echoProperties();
	
	// Print out header line "p cnf..."
	isset($NVARS) or die("maxVar is not set");
	echo("p cnf $NVARS $NCLAUSES\n");
	
	
	// Print out all clauses
	echoClauses();	

	//print_r($dictionary);
	
	


	function addClause($clause) {
		global $CLAUSES,$NCLAUSES;
		$clause = trim($clause);
		if ($clause == "") return;
		$clause .= " 0";
		$CLAUSES[] = $clause; // Append to array
		$NCLAUSES++;
	}

	function addComment($msg) {
		global $CLAUSES;
		$CLAUSES[] = $msg;
		// No $NCLAUSES increment here!
	}


	function plays2dimacs($p,$g,$w) {
		global $NVARS, $dictionary, $DEBUG, $P, $G, $W;
		
		if ($p < 1 || $p > $P) {
			echoClauses();
			die("Trying to retrieve a 'PLAYS' variable for player $p: PLAYS($p,$g,$w)");
		}
		if ($g < 1 || $g > $G) {
			echoClauses();		
			die("Trying to retrieve a 'PLAYS' variable for group $g: PLAYS($p,$g,$w)");
		}
		if ($w < 1 || $w > $W) {
			echoClauses();		
			die("Trying to retrieve a 'PLAYS' variable for week $w: PLAYS($p,$g,$w)");
		}
		$result = "PLAYS(p=$p,g=$g,w=$w)";		
		
		$v = $dictionary['PLAYS'][$p][$g][$w];
		if ($v !== NULL) {
			if (!$DEBUG) $result = $v;
		}
		else {
			$NVARS++;
			$dictionary['PLAYS'][$p][$g][$w] = $NVARS;
			if (!$DEBUG) $result = $NVARS;
		}
		return $result;
	}
	
	
	function allingroup2dimacs($g,$w,$tuple) {
		global $NVARS, $dictionary, $DEBUG, $P, $G, $W;
		
		if ($g < 1 || $g > $G) {
			echoClauses();		
			die("Trying to retrieve a 'PLAYS' variable for group $g: PLAYS($p,$g,$w)");
		}
		if ($w < 1 || $w > $W) {
			echoClauses();		
			die("Trying to retrieve a 'PLAYS' variable for week $w: PLAYS($p,$g,$w)");
		}

		$tupleString = join(",",$tuple);
		
		$result = "ALL_IN_GROUP(g=$g,w=$w,tuple=<".$tupleString.">)";
		
		$v = $dictionary['ALL_IN_GROUP'][$g][$w][$tupleString];
		if ($v != NULL) {
			if (!$DEBUG) $result = $v;
		}
		else {
			$NVARS++;
			$dictionary['ALL_IN_GROUP'][$g][$w][$tupleString] = $NVARS;
			if (!$DEBUG) $result = $NVARS;
		}
		return $result;
	}	
	
	
	
	function echoClauses() {
		global $CLAUSES;
		foreach($CLAUSES as $c) {
			echo($c ."\n");
		}
	}	
	
	function echoDictionary() {
		global $DEBUG, $dictionary;
		echo ("c\nc START DICTIONARY\n");
		if ($DEBUG) {
			echo("c omitted -- DEBUG mode\n");
		}
		else {		
			if (isset($dictionary['PLAYS'])) {			
				foreach($dictionary['PLAYS'] as $p => $entryP) {
					foreach($entryP as $g => $entryG) {
						foreach($entryG as $w => $dimacsVar) {
							echo("c plays($p,$g,$w) = $dimacsVar\n");
						}
					}
				}
			}		
			if (isset($dictionary['ALL_IN_GROUP'])) {
				foreach($dictionary['ALL_IN_GROUP'] as $g => $entryG) {
						foreach($entryG as $w => $entryW) {
							foreach($entryW as $tuple => $dimacsVar) {
								echo("c all_in_group($g,$w,<$tuple>) = $dimacsVar\n");
							}
						}
				}	
			}
		}
		echo("c END DICTIONARY\nc\nc\n");		
	}	
	
	function echoProperties() {
		global $DEBUG, $dictionary;	
		// All_IN_GROUP variables are dependent on some PLAYS variables
		if (isset($dictionary['ALL_IN_GROUP'])) {
			echo("c START PROPERTIES\n");
			foreach($dictionary['ALL_IN_GROUP'] as $g => $entryG) {
					foreach($entryG as $w => $entryW) {
						foreach($entryW as $tuple => $dimacsVar) {
							// ALL_IN_GROUP($g,$w,<$p1...$ps>) DEPENDENT ON PLAYS_p1_g_w ... PLAYS_ps_g_w
							$tupleArray = explode(",", $tuple);							
							echo("c " . allingroup2dimacs($g,$w,$tupleArray). " DEPENDENT ON ");
							foreach($tupleArray as $t) {
								echo( plays2dimacs($t,$g,$w) . " ");
							}
							echo("\n");
						}
					}
			}	
			echo("c END PROPERTIES\n");
		}
	}
	
	
	
	

	function firstTuple() {
		global $P, $S;
		$result = array();
		for ($i = 0; $i<$S; $i++) {
			$result[$i] = $i+1;
		}
		return $result;
	}

	function nextTuple($tuple) {
		global $P, $S;
		$result = $tuple; //array();
		/*foreach($tuple as $k=>$v) { 
			$result[$k] = $v; 
		}*/		
		
		for ($i=$S-1; $i>=0; $i--) {		
			$remaining = $S-1 - $i;
			if ($result[$i] < $P && ($result[$i]+1 <= $P-$remaining )   ) {
				$result[$i]++;			
				for ($j = $i+1; $j < $S; $j++) {
					$result[$j] = $result[$i] + ($j - $i);
				}			
				return $result;				
			}
		}
		return NULL;
	}
?>